@@ -1,3 +1,65 @@
+"""
+CheckedDict
+-----------
+
+A dictionary based on a default prototype. A CheckedDict can only define
+``key:value`` pairs which are already present in the default. It is possible to
+define a docstring for each key and different restrictions for the values
+regarding possible values, ranges and type. A CheckedDict is useful for
+configuration settings.
+
+
+ConfigDict
+----------
+
+Based on CheckedDict, a ConfigDict is a persistent, unique dictionary. It is
+saved under the config folder determined by the OS and it is updated with each
+modification. It is useful for implementing configuration of a module / library
+/ app, where there is a default/initial state and the user needs to be able to
+configure global settings which must be persisted between sessions (similar to
+the settings in an application)
+
+Example::
+
+
+    from configdict import ConfigDict
+
+    config = ConfigDict("myproj.subproj")
+    config.addKey("keyA", 10, doc="documentaion of keyA")
+    config.addKey("keyB", 0.5, range=(0, 1))
+    config.addKey("keyC", "blue", choices=("blue", "red"), doc="documentation of keyC")
+    config.load()
+
+
+Alternatively, a ConfigDict can be created all at once::
+
+
+    config = ConfigDict("myapp",
+        default = {
+            'font-size': 10.0,
+            'font-family': "Monospace",
+            'port' : 9100,
+        },
+        validator = {
+            'font-size::range' : (8, 24),
+            'port::range' : (9000, 65000),
+            'font-family::choices' : {'Roboto', 'Monospace'},
+            'port': lambda cfg, port: checkPortAvailable(port)
+        },
+        docs = {
+            'port': 'The port number to listen to',
+            'font-size': 'The size of the font, in pixels'
+        }
+    )
+
+
+This will create the dictionary and load any persisted version. Any saved
+modifications will override the default values. Whenever the user changes any
+value (via ``config[key] = newvalue``) the dictionary will be saved.
+
+In all other respects a ConfigDict behaves like a normal dictionary.
+
+"""
 from __future__ import annotations
 
 import appdirs
@@ -10,13 +72,16 @@ import re
 import textwrap
 import tempfile
 from types import FunctionType
-from typing import (Optional as Opt, Any, Tuple, Dict, Union)
+from typing import (Optional as Opt, Any, Tuple, Dict, Union, Callable, TypeVar)
 
 
 __all__ = ["CheckedDict", "ConfigDict", "getConfig", "activeConfigs", "configPathFromName"]
 
 logger = logging.getLogger("configdict")
 
+validatefunc_t = Callable[[dict, Any], bool]
+T = TypeVar("T")
+
 _UNKNOWN = object()
 
 
@@ -212,39 +277,40 @@ def _openInEditor(cfg):
     _openInStandardApp(cfg)
 
 
-
 class CheckedDict(dict):
+    """
+    A dictionary which checks that the keys and values are valid
+    according to a default dict and a validator.
+
+    Args:
+        default: a dict will all default values. A config can accept only
+            keys which are already present in the default
+
+        validator: a dict containing choices and types for the keys in the
+            default. Given a default like: ``{'keyA': 'foo', 'keyB': 20, 'keyC': 0.5}``,
+            a validator could be::
+
+                {'keyA::choices': ['foo', 'bar'],
+                 'keyB::type': float,
+                 'keyB': lambda d, value: value > d['keyC'] * 10
+                 'keyC::range': (0, 1)
+                }
+
+            choices can be defined lazyly by giving a lambda which returns a list
+            of possible choices
+
+        docs: a dict containing help lines for keys defined in default
+        callback: function ``(key, value) -> None``. This function is called **after**
+            the modification has been done.
+        precallback: function ``(key, value) -> None``
+    """
     def __init__(self,
                  default: Dict[str, Any] = None,
                  validator: Dict[str, Any] = None,
                  docs: Dict[str, str] = None,
-                 callback=None,
+                 callback:Callable[[str, Any], None]=None,
                  precallback=None) -> None:
-        """
-        A dictionary which checks that the keys and values are valid
-        according to a default dict and a validator.
 
-        Args:
-            default: a dict will all default values. A config can accept only
-                keys which are already present in the default
-
-            validator: a dict containing choices and types for the keys in the
-                default. Given a default like: {'keyA': 'foo', 'keyB': 20},
-                a validator could be:
-
-                {'keyA::choices': ['foo', 'bar'],
-                 'keyB::type': float,
-                 'keyC::range': (0, 1)
-                }
-
-                choices can be defined lazyly by giving a lambda which returns a list
-                of possible choices
-
-            docs: a dict containing help lines for keys defined in default
-            callback:
-                function (key, value) -> None
-                This function is called AFTER the modification has been done.
-        """
         self.default = default if default else {}
         self._validator = _checkValidator(validator,
                                           default) if validator else {}
@@ -257,6 +323,9 @@ class CheckedDict(dict):
         self._allowedkeys = set(self.default.keys())
 
     def copy(self) -> CheckedDict:
+        """
+        Create a copy of this CheckedDict
+        """
         out = CheckedDict(default=self.default, validator=self._validator, docs=self._docs,
                           precallback=self._precallback, callback=self._callback)
         return out
@@ -275,21 +344,27 @@ class CheckedDict(dict):
 
     def addKey(self,
                key: str,
-               value,
-               type=None,
+               value: Any,
+               type: Union[type, Tuple[type,...]]=None,
                choices=None,
                range: Tuple[Any, Any] = None,
+               validatefunc: validatefunc_t = None,
                doc: str = None) -> None:
         """
-        Add a key: value pair to the default settings. This is used when building the
+        Add a ``key: value`` pair to the default settings. This is used when building the
         default config item by item (see example). After adding all new keys it is
-        necessary to call .load()
+        necessary to call ``.load()``
+
+        Example::
+
 
-        Example:
             cfg = ConfigDict("foo", load=False)
             # We define a default step by step
-            cfg.addKey("size", 100, range=(50, 150))
+            cfg.addKey("width", 100, range=(50, 150))
             cfg.addKey("color", "red", choices=("read", "blue", "green"))
+            cfg.addKey("height",
+                       doc="Height should be higher than width",
+                       validatefunc=lambda cfg, height: height > cfg['width'])
             # Now update the dict with the newly defined default and any
             # saved version
             cfg.load()
@@ -299,7 +374,9 @@ class CheckedDict(dict):
             value: a default value
             type: the type accepted, as passed to isinstance (can be a tuple)
             choices: a seq of possible values
-            range: a (min, max) tuple defining allowed range
+            range: a (min, max) tuple defining an allowed range for this value
+            validatefunc: a function ``(config, value) -> bool``, which should return
+                `True` if value is valid for `key` or False otherwise
             doc: documentation for this key
 
         """
@@ -312,6 +389,9 @@ class CheckedDict(dict):
             validator[f"{key}::choices"] = choices
         if range:
             validator[f"{key}::range"] = range
+        if validatefunc:
+            assert callable(validatefunc)
+            validator[key] = validatefunc
         if doc:
             self._docs[key] = doc
 
@@ -335,6 +415,17 @@ class CheckedDict(dict):
             self._callback(key, value)
 
     def checkDict(self, d: dict) -> str:
+        """
+        Check if dict `d` can be used to update self
+
+        Args:
+            d (dict): a dict which might update self
+
+        Returns:
+            An error message if `d` has any invalid `key` or `value`,
+            "" if everything is ok
+
+        """
         invalidkeys = [key for key in d if key not in self.default]
         if invalidkeys:
             return f"Some keys are not valid: {invalidkeys}"
@@ -344,10 +435,26 @@ class CheckedDict(dict):
                 return errormsg
         return ""
 
+    def getValidateFunc(self, key:str) -> Opt[validatefunc_t]:
+        """
+        Returns a function to validate a value for `key`, if there
+        is one. A validate function has the form ``(config, value) -> bool``
+
+        Args:
+            key (str): the key to query for a validate function
+
+        Returns:
+            The validate function, or None
+
+        """
+        func = self._validator(key, None)
+        assert func is None or callable(func)
+        return func
+
     def getChoices(self, key: str) -> Opt[list]:
         """
         Return a seq. of possible values for key `k`
-        or None
+        or `None`
         """
         if key not in self._allowedkeys:
             raise KeyError(f"{key} is not a valid key")
@@ -373,15 +480,20 @@ class CheckedDict(dict):
 
         Returns errormsg. If value is of correct type, errormsg is None
 
-        Example:
+        Example::
 
-        error = config.checkType(key, value)
-        if error:
-            print(error)
+            error = config.checkType(key, value)
+            if error:
+                print(error)
         """
         choices = self.getChoices(key)
         if choices is not None and value not in choices:
             return f"key {key} should be one of {choices}, got {value}"
+        func = self.getValidateFunc(key)
+        if func:
+            ok = func(self, value)
+            if not ok:
+                return f"{value} is not valid for key {key}"
         t = self.getType(key)
         if t == float:
             if not _isfloaty(value):
@@ -396,22 +508,27 @@ class CheckedDict(dict):
         return None
 
     def getRange(self, key: str) -> Opt[tuple]:
+        """
+        Returns the valid range for the value corresponding to this key,
+        if it was specified.
+        """
         if key not in self._allowedkeys:
             raise KeyError(f"{key} is not a valid key")
         if not self._validator:
-            logger.debug("getChoices: validator not set")
+            logger.debug("getRange: validator not set")
             return None
         return self._validator.get(key+"::range", None)
 
-    def getType(self, key: str):
+    def getType(self, key: str) -> Union[type, Tuple[type,...]]:
         """
-        Returns the expected type for key, as a type
+        Returns the expected type for key, as a type which can be passed
+        to isinstance
 
-        NB: all numbers are reduced to type float, all strings are of type str,
+        **NB**: all numbers are reduced to type float, all strings are of type str,
             otherwise the type of the default value, which can be a collection
             like a list or a dict
 
-        See Also: checkValue
+        See Also: `checkValue`
         """
         if self._validator is not None:
             definedtype = self._validator.get(key+"::type")
@@ -431,6 +548,11 @@ class CheckedDict(dict):
                                  (bytes, str)) else type(defaultval)
 
     def getTypestr(self, key: str) -> str:
+        """
+        The same as `.getType` but returns a string representation of the type/types
+        possible for the value of this key
+
+        """
         t = self.getType(key)
         if isinstance(t, tuple):
             return "("+", ".join(x.__name__ for x in t)+")"
@@ -440,17 +562,14 @@ class CheckedDict(dict):
     def reset(self) -> None:
         """
         Resets the config to its default (inplace), and saves it.
-
-        Example
-        ~~~~~~~
-
-        cfg = getconfig("folder:config")
-        cfg = cfg.reset()
         """
         self.clear()
         self.update(self.default)
 
     def update(self, d: dict=None, **kws) -> None:
+        """
+        Update ths dict with `d` or any key:value pair passed as keyword
+        """
         if d:
             errormsg = self.checkDict(d)
             if errormsg:
@@ -462,52 +581,50 @@ class CheckedDict(dict):
                 raise ValueError(f"invalid keywords: {errormsg}")
             super().update(kws)
 
-    def override(self, key: str, value, default=None):
+    def updated(self:T, d: dict=None, **kws) -> T:
         """
-        The same as `value if value is not None else config.get(key, default)
+        The same as :meth:`~CheckedDict.update`, but returns self after the operation
+        """
+        self.update(d, **kws)
+        return self
+
+    def override(self, key: str, value, default=None) -> None:
+        """
+        The same as `value if value is not None else config.get(key, default)`
         """
         return value if value is not None else self.get(key, default)
 
     def asYaml(self) -> str:
+        """
+        Returns this dict as yaml str, with comments, defaults, etc.
+        """
         return _asYaml(self, doc=self._docs, validator=self._validator,
                        default=self.default)
 
 
 class ConfigDict(CheckedDict):
-    registry: Dict[str, ConfigDict] = {}
-    _helpwidth: int = 58
+    """
+    This is a (optionally) persistent, unique dictionary used for configuration
+    of a module / app. It is saved under the config folder determined by
+    the OS (and is thus OS dependent) and no two instances of the same
+    config can coexist.
 
-    def __init__(self,
-                 name: str,
-                 default: Dict[str, Any] = None,
-                 validator: Dict[str, Any] = None,
-                 docs: Dict[str, str] = None,
-                 precallback=None,
-                 persistent=True,
-                 load=True,
-                 fmt='yaml') -> None:
-        """
-        This is a (optionally) persistent, unique dictionary used for configuration
-        of a module / app. It is saved under the config folder determined by
-        the OS (and is thus OS dependent) and no two instances of the same
-        config can coexist.
+    Args:
+        name: a str of the form ``prefix.name`` or ``prefix/name``
+            (these are the same) or simply ``name`` if this is an
+            isolated configuration (not part of a bigger project). The
+            data will be saved at ``$USERCONFIGDIR/{prefix}/{name}.{fmt}`` if
+            prefix is given, or ``$USERCONFIGDIR/{name}.{fmt}``.
+            For instance, in Linux a config with a name "myproj.myconfig" and
+            a yaml format will be saved to "~/.config/mydir/myconfig.yaml"
 
-        Args:
-            name: a str of the form ``prefix.name`` or ``prefix/name``
-                (these are the same) or simply ``name`` if this is an
-                isolated configuration (not part of a bigger project). The
-                data will be saved at ``$USERCONFIGDIR/{prefix}/{name}.{fmt}`` if 
-                prefix is given, or ``$USERCONFIGDIR/{name}.{fmt}``.
-                For instance, in Linux a config with a name "myproj.myconfig" and 
-                a yaml format will be saved to "~/.config/mydir/myconfig.yaml" 
+        default: a dict with all default values. A config can accept only
+            keys which are already present in the default. This value can be
+            left as None if the config is built successively. See example below
 
-            default: a dict with all default values. A config can accept only
-                keys which are already present in the default. This value can be
-                left as None if the config is built successively. See example below
-
-            validator: a dict containing choices, types and/or ranges for the keys in the
-                default. Given a default like: ``{'keyA': 'foo', 'keyB': 20}``,
-                a validator could be:
+        validator: a dict containing choices, types and/or ranges for the keys in the
+            default. Given a default like: ``{'keyA': 'foo', 'keyB': 20}``,
+            a validator could be::
 
                 {
                   'keyA::choices': ['foo', 'bar'],
@@ -515,63 +632,77 @@ class ConfigDict(CheckedDict):
                   'keyB::range': (10, 30)
                 }
 
-                Choices can be defined lazyly by giving a lambda
-                
-            docs: a dict containing documentation for each key
-            
-            persistent: if True, any change to the dict will be saved. 
-            
-            load: if True, the saved version will be loaded after creation. This is disabled if 
-                no default dict is given. .load should be called manually in this case (see example)
 
-            precallback:
-                function (self, key, oldvalue, newvalue) -> None|newvalue,
-                If given, it is called BEFORE the modification is done. This function
-                should return:
-                    * None to allow modification
-                    * any value to modify the value
-                    * raise a ValueError exception to stop the transaction
+            Choices can be defined lazyly by giving a lambda
 
-        Example:
-            A ConfigDict defined item by item
+        docs: a dict containing documentation for each key
 
-            config = ConfigDict("myproj.subproj")
-            config.addKey("keyA", 10, doc="documentaion of keyA")
-            config.addKey("keyB", 0.5, range=(0, 1))
-            config.addKey("keyC", "blue", choices=("blue", "red"),
-                          doc="documentation of keyC")
-            config.load()
+        persistent: if True, any change to the dict will be saved.
 
-            # The same effect can be achieved by passing the default/validator/doc 
+        load: if True, the saved version will be loaded after creation. This is disabled if
+            no default dict is given. .load should be called manually in this case (see example)
 
-            default = {
-                "keyA": 10,
-                "keyB": 0.5,
-                "keyC": "blue
-            }
+        precallback: function `(dict, key, oldvalue, newvalue) -> None|newvalue`,
+            If given, it is called *before* the modification is done. This function
+            should return **None** to allow modification, **any value** to modify the value, or
+            **raise ValueError** to stop the transaction
 
-            validator = {
-                "keyB::range" = (0, 1),
-                "keyC::choices" = ("blue", "red")
-            }
 
-            docs = {
-                "keyA": "documentation of keyA"
-                "keyC": "documentation of keyC"
-            }
+    Example::
 
-            cfg = ConfigDict("myproj.subproj",
-                             default=default,
-                             validator=validator,
-                             docs=docs)
-            # no need to call .load in this case
 
-        """
+        config = ConfigDict("myproj.subproj")
+        config.addKey("keyA", 10, doc="documentaion of keyA")
+        config.addKey("keyB", 0.5, range=(0, 1))
+        config.addKey("keyC", "blue", choices=("blue", "red"),
+                      doc="documentation of keyC")
+        config.load()
+
+        # The same effect can be achieved by passing the default/validator/doc
+
+        default = {
+            "keyA": 10,
+            "keyB": 0.5,
+            "keyC": "blue
+        }
+
+        validator = {
+            "keyB::range": (0, 1),
+            "keyC::choices": ("blue", "red")
+        }
+
+        docs = {
+            "keyA": "documentation of keyA"
+            "keyC": "documentation of keyC"
+        }
+
+        cfg = ConfigDict("myproj.subproj",
+                         default=default,
+                         validator=validator,
+                         docs=docs)
+        # no need to call .load in this case
+
+    """
+
+    _registry: Dict[str, ConfigDict] = {}
+
+    _helpwidth: int = 58
+
+    def __init__(self,
+                 name: str,
+                 default: Dict[str, Any] = None,
+                 validator: Dict[str, Any] = None,
+                 docs: Dict[str, str] = None,
+                 precallback:Callable[[ConfigDict, str, Any, Any], Any]=None,
+                 persistent=True,
+                 load=True,
+                 fmt='yaml') -> None:
+
         if name:
             name = _normalizeName(name)
             if not _isValidName(name):
                 raise ValueError(f"name {name} is invalid for a config")
-        if name in ConfigDict.registry:
+        if name in ConfigDict._registry:
             logger.warning("A ConfigDict with the given name already exists!")
         self.fmt = fmt
 
@@ -604,6 +735,10 @@ class ConfigDict(CheckedDict):
 
     @property
     def name(self) -> Opt[str]:
+        """
+        The name of this ConfigDict. The name determines where it is saved
+        (if persist==True)
+        """
         return self._name
 
     @name.setter
@@ -611,15 +746,16 @@ class ConfigDict(CheckedDict):
         if self._name:
             raise ValueError("Name has already been set")
 
-        if name and name in self.registry:
+        if name and name in self._registry:
             raise ValueError(f"Name {name} is already used")
         self._name = name
         base, configname = _parseName(name)
         self._base: str = base
-        self.registry[name] = self
+        self._registry[name] = self
 
     @property
     def persistent(self) -> bool:
+        """Is this a persistent ConfigDict?"""
         return self._persistent
 
     @persistent.setter
@@ -677,7 +813,7 @@ class ConfigDict(CheckedDict):
         """
         return self.clone(name='', persistent=False, cloneCallbacks=False)
 
-    def clone(self, name: str = '', persistent: bool=None, cloneCallbacks=False
+    def clone(self, name: str = '', persistent: bool=None, cloneCallbacks=False,
               ) -> ConfigDict:
         """
         Create a clone of this dict
@@ -693,7 +829,7 @@ class ConfigDict(CheckedDict):
         Returns:
             the cloned dict
         """
-        if name == self._name or name in self.registry:
+        if name == self._name or name in self._registry:
             raise ValueError(f"name {name} is already taken!")
         out = ConfigDict(default=self.default, validator=self._validator, docs=self._docs,
                          persistent=False, load=False, name=name)
@@ -705,18 +841,19 @@ class ConfigDict(CheckedDict):
                 out.registerCallback(func, pattern)
         return out
 
-    def registerCallback(self, func, pattern=None) -> None:
+    def registerCallback(self, func:Callable[[ConfigDict, str, Any], None], pattern:str=None) -> None:
         """
         Register a callback to be fired when a key matching the given pattern is
         changed. If no pattern is given, the function will be called for
         every key.
 
         Args:
-            func: a function of the form (dict, key, value) -> None
-                dict - this ConfigDict itself
-                key - the key which was just changed
-                value - the new value
-            pattern: call func when pattern matches key
+            func: a function of the form ``(dict, key, value) -> None``, where *dict* is
+                this ConfigDict itself, *key* is the key which was just changed and *value*
+                is the new value.
+
+            pattern: call func when pattern matches key.
+
         """
         self._callbacks.append((pattern or r".*", func))
 
@@ -735,6 +872,11 @@ class ConfigDict(CheckedDict):
         """
         Normally a config doesn't need to be saved by the user,
         it is saved whenever it is modified.
+
+        Args:
+            path (str): the path to save the config. If None and this
+                is a named config, it is saved to the path returned by
+                :meth:`~ConfigDict.getPath`
         """
         if path is None:
             path = self.getPath()
@@ -798,7 +940,7 @@ class ConfigDict(CheckedDict):
     def edit(self, waitOnModified=False) -> None:
         """
         Edit this config by opening it in an external application. The format
-        used is yaml, because it allows to embed comments. This is independent
+        used is *yaml*, because it allows to embed comments. This is independent
         of the format used for persistence. The application used is the user's
         default application for the .yaml format and can be configured at the
         os level. In macos we use `open`, in linux `xdg-open` and in windows
@@ -828,15 +970,7 @@ class ConfigDict(CheckedDict):
 
     def load(self, configpath:str=None) -> None:
         """
-        Read the saved config, update self. This is used internally but it can be usedful
-        if the file is changed externally and no monitoring is activated
-
-        * If no saved config (not present or unreadable)
-            * if default was given:
-                * use default
-            * otherwise:
-                * if saved config is unreadable, raise JSONDecodeError
-                * if saved config not present, raise FileNotFoundError
+        Read the saved config, update self.
         """
         if configpath is None:
             configpath = self.getPath()
@@ -906,7 +1040,7 @@ def _mergeDicts(readdict: Dict[str, Any], default: Dict[str, Any]) -> Dict[str,
         default:
 
     Returns:
-
+        the merged dict
     """
     out = {}
     sharedkeys = readdict.keys() & default.keys()
@@ -976,14 +1110,14 @@ def getConfig(name: str) -> Opt[ConfigDict]:
     """
     name = _normalizeName(name)
     _checkName(name)
-    return ConfigDict.registry.get(name)
+    return ConfigDict._registry.get(name)
 
 
 def activeConfigs() -> Dict[str, ConfigDict]:
     """
     Returns a dict of active configs
     """
-    return ConfigDict.registry.copy()
+    return ConfigDict._registry.copy()
 
 
 def _removeConfigFromDisk(name: str) -> bool:
